import express from "express";
import fetch from "sync-fetch";
import chalk from "chalk";
import * as fs from 'fs';
import { NodeVM } from 'vm2';
// import { createRequire } from "module";
// const require = createRequire(
//     import.meta.url);

/* performance */
// import sizeof from 'object-sizeof'
// var datasize = {};
const timeperformance = false; 

/* search for [BENCHMARK] to find out possible executions of different cases */
/* to see the benchmark evaluation table: https://docs.google.com/spreadsheets/d/1aNXyPmBDZEWjKHD8cGTQfCiTkVT57zUaRvzooJqF7PI/edit?usp=sharing */

// TAP
const tap = express();
tap.use(express.json());

// query types

// all proxy, querytypes just for optimization
// const querytypes = require("./querytypes.json");
// for node version 17 and above!
import querytypes from './querytypes.json' assert {type: 'json'};

// query functions 
import * as QueryFunc from './queryFunctions.js'

// [BENCHMARK] note that the trigger/user data is shared between cases
// Meta object (should be a query?)
var config = {
    currentTime: "2020-04-23T18:25:43.511Z",
    timeZone: "GMT",
    triggerTime: "2020-04-23T19:25:43.511Z"
}
import moment from "moment-timezone"
var Meta = {
    currentUserTime: moment(config.currentTime).tz(config.timeZone)
}
Meta.triggerTime = config.triggerTime === null ? null : moment(config.triggerTime).tz(config.timeZone)

// atttribute counter
var fetchedAttributes = [];


// type - the type of event
// id - the id of the specific event
tap.post('/trigger/:servicename/:type', (req, res) => {

    //performance
    timeStart("trigger-to-action");

    let servicename = req.params.servicename;
    let type = req.params.type;
    let id = req.body.id;

    const vm = new NodeVM({
        console: 'inherit', //default
        sandbox: {
            log,
            servicename,
            type,
            id,
            LazyService,
            RemoteObject,
            RemoteArray,
            // LazyProjection,
            Strictify,
            QueryFunc,
            config,
            Meta        
        },
        // sandbox: { log, servicename, type, id, LazyService, buildLazy, QueryFunc, config, Meta },
        eval: false, //default : true
        require: {
            external: false, //default
            builtin: [], //default
            root: "./",
            mock: {}
        }
    });
    vm.freeze(moment, 'moment');



    log(`from service ${servicename} event type ${type} with id ${id} received`);

    switch (id) {
        //designed cases
        case "001ets":
            vm.run(fs.readFileSync('./designed-cases/1/'.concat(id + ".js"), 'utf8'));
            break;

        case "002tcs":
            vm.run(fs.readFileSync('./designed-cases/2/'.concat(id + ".js"), 'utf8'));
            break;

        case '003ecs':
            vm.run(fs.readFileSync('./designed-cases/3/'.concat(id + ".js"), 'utf8'));
            break;

        case '004scpn':
            vm.run(fs.readFileSync('./designed-cases/4/'.concat(id + ".js"), 'utf8'));
            break;

        case '005ecfsg':
            vm.run(fs.readFileSync('./designed-cases/5/'.concat(id + ".js"), 'utf8'));
            break;

        case '006tyt':
            vm.run(fs.readFileSync('./designed-cases/6/'.concat(id + ".js"), 'utf8'));
            break;




            //ifttt cases
        case 'AxJSC34d':
            vm.run(fs.readFileSync('./from-ifttt-dataset/1/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'example-03':
            vm.run(fs.readFileSync('./from-ifttt-dataset/2/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'iqZPNUtR':
            vm.run(fs.readFileSync('./from-ifttt-dataset/3/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'jUy5if7H':
            vm.run(fs.readFileSync('./from-ifttt-dataset/4/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'MRm9VBxG':
            vm.run(fs.readFileSync('./from-ifttt-dataset/5/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'YG5HSLvK':
            vm.run(fs.readFileSync('./from-ifttt-dataset/6/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'xvyUBQsh':
            vm.run(fs.readFileSync('./from-ifttt-dataset/7/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'example-04':
            vm.run(fs.readFileSync('./from-ifttt-dataset/8/'.concat(id + "-appcode.js"), 'utf8'));
            break;

        case 'VDdNBmiE':
            vm.run(fs.readFileSync('./from-ifttt-dataset/9/'.concat(id + "-appcode.js"), 'utf8'));
            break;





            //skeleton
        case 'skeleton':
            log("##### example XX is running... #####")

            //preamble 
            //trigger

            //query

            //action

            // filter code
            log("** filter code **")

            log("** end of filter code **")

            // post app action 

            //result
            log("action: ", Action)
            log("##### end of example XX #####")

            break;

        default:
            log('error: no apps found!')
            break;
    }

    // performance
    timeEnd("trigger-to-action");


    console.log('\n\n\n')
    log("list of fetched attributes:\n" + chalk.blue(fetchedAttributes.join('\n')) + '\n' + '#attributes: ' + chalk.blue(fetchedAttributes.length) + '\n')

    // performance
    // log("Size of transferred data", datasize)

    res.json(""); // to make postJSON happy
    res.end();

});

tap.listen(9000);



// LazyService is both LazyQuery and LazyTrigger in the formalization

export function Strictify(value) {
    // if (value instanceof LazyProjection) return value.Project();
    if (value instanceof Function) return value();
    return value;
}


// export class LazyProjection {
//     constructor(obj /* remote, never local */ , prop /* not remote */ ) {
//         this.obj = obj;
//         this.prop = prop;
//     }

//     Project() {
//         this.obj = Strictify(this.obj);
//         return this.obj[this.prop];
//     }

// }



export class LazyService {
    constructor(url, args) {
        this.url = url;
        this.args = args;
        // this.nameInUrl = this.url.replace('http://localhost:4000/', '');
        // datasize[this.nameInUrl] = 0;
    }

    fetch(path) {
        if (this.id === undefined) { // lazy service has been set up?
            // seq args
            for (let p in this.args) {
                this.args[p] = Strictify(this.args[p]);
            }
            log(`setting up query with url ${this.url} and arguments ${JSON.stringify(this.args)}`);

            let result = postJSON(this.url, this.args);
            this.id = result.id;
        }

        log(chalk.black(`fetching result for ${this.url}/${this.id}${path}`))

        // aggregating fetched attributes to report at the end of the execution
        let attribute = this.url.replace('http://localhost:4000/', '');
        attribute += `/${this.id}${path}`
        attribute = attribute.split('/');
        attribute.splice(2, 1);
        fetchedAttributes.push(attribute.join('.'))
        //


        let result = postJSON(`${this.url}/${this.id}${path}`);
        
        //performance
        // datasize[this.nameInUrl] += sizeof(result)

        return result;
    }
}
            


export class RemoteObject {

    constructor(service, path, structure) {
        this.service = service; // fetcher
        this.path = path;
        this.structure = structure;

        this.handler = {
            get: this.get.bind(this),
            has: this.has.bind(this),
            ownKeys: this.ownKeys.bind(this),
            getOwnPropertyDescriptor: this.getOwnPropertyDescriptor.bind(this),
            set: this.unsupported.bind(this, "set"),
            deleteProperty: this.unsupported.bind(this, "deleteProperty"),
            defineProperty: this.unsupported.bind(this, "defineProperty")
        };

    }

    static Create(service, path = "", structure = querytypes[service.url.split('/')[3]][service.url.split('/')[4]]) {
        let obj = new RemoteObject(service, path, structure);
        return new Proxy({} /* ref */ , obj.handler);
    }

    get(target, prop) {
        if (target[prop] !== undefined) {
            return target[prop];
        }
        var fullPath = this.path + "/" + prop;

        // structure communication optimization
        var struct = this.structure;
        for (let p of fullPath.split("/")) {
            if (struct instanceof Array) struct = struct[0];
            else if (struct instanceof Object) {
                if (p === '') continue; //zeroth index, not array
                else if (struct[p] !== undefined) struct = struct[p];
                else if (!/^\d+$/.test(p)) //not an array index
                    throw fullPath + ` doesn't comply with the structure` + JSON.stringify(this.structure)
            }
        }


        // prim, {}, []
        if (struct instanceof Array) {
            target[prop] = RemoteArray.Create(this.service, fullPath);
        } else if (struct instanceof Object) {
            target[prop] = RemoteObject.Create(this.service, fullPath);
        } else {
            var value = this.service.fetch(fullPath);
            if (value !== undefined) target[prop] = value;
        }

        return target[prop];
    }

    has(target, prop) {
        this.get(target, prop);
        return Object.hasOwn(target, prop);
    }

    ownKeys(target) {
        // throw 'ownKeys not supported!'
        return Reflect.ownKeys(target);
    }

    getOwnPropertyDescriptor(target, key) {
        // throw 'getOwnPropertyDescriptor not supported!'
        return Reflect.getOwnPropertyDescriptor(target, key)
    }

    unsupported(prop) {
        throw prop + ' is not allowed for remote proxied objects!'
    }
}

export class RemoteArray extends RemoteObject {
    constructor(service, path, structure) {
        super(service, path, structure);
        this.firstLength = true;
    }

    static Create(service, path = "", structure = querytypes[service.url.split('/')[3]][service.url.split('/')[4]]) {
        let arr = new RemoteArray(service, path, structure);
        return new Proxy([] /* ref */ , arr.handler);
    }

    get(target, prop) {
        if (prop == "length") {
            // we must handle length differently, since it exists on target
            if (this.firstLength) {
                target[prop] = this.service.fetch(this.path + "/" + prop);
                this.firstLength = false;
            }
        }
        return super.get(target, prop);

    }
}

// ---

function postJSON(url, data) {
    let result = fetch(url, {
        method: "POST",
        headers: {
            'Content-Type': 'application/json'
        },
        body: JSON.stringify(data)
    });

    return result.json();
}

function timeStart(){
    console.time.apply(console, arguments);
}

function timeEnd(){
    process.stdout.write(chalk.magenta("[LazyTAP] "));
    console.timeEnd.apply(console, arguments);
}

function log() {
    if (timeperformance) return; 
    process.stdout.write("[LazyTAP] ");
    console.log.apply(console, arguments);
}